
code {
        background-color: #eeeeee;
	text-shadow: none;
	color: black;
	margin-left: 4px;
	margin-right: 4px;
	padding-left: 4px;
	padding-right: 4px;
	border-radius: 3px;
	white-space: nowrap;
}
